#include "ofxUvcCamera.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <fcntl.h>              /* low-level i/o */
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <linux/videodev2.h>
#include <linux/uvcvideo.h>
#include <linux/usb/video.h>
#include <linux/usbdevice_fs.h>

#define CLEAR(x) memset(&(x), 0, sizeof(x))

static void errno_exit(const char *s)
{
        fprintf(stderr, "%s error %d, %s\n", s, errno, strerror(errno));
        exit(EXIT_FAILURE);
}

int ofxUvcCamera::xioctl(int fh, int request, void *arg)
{
        int r;

        do {
                r = ioctl(fh, request, arg);
        } while (-1 == r && EINTR == errno);

        return r;
}

int ofxUvcCamera::setDepthType()
{
	unsigned char SetupPacket[64] = {0};
	struct uvc_xu_control_query xqry;

	int r=0;
	*(SetupPacket)=0x20;
	*(SetupPacket+1)=0xf0; 
	*(SetupPacket+2)=0x02; // 0x02 14Bit Depth , 0x04 11Bit Depth

	xqry.unit = 0x04;
	xqry.selector = 3; //0x07;
	xqry.size =4;
	xqry.data = (__u8*)malloc(64);
	memcpy(xqry.data,SetupPacket,59);

	xqry.query = UVC_SET_CUR;
#if 1
	if  ( (r = xioctl(fd,UVCIOC_CTRL_QUERY,&xqry))<0 )
	{
		if ( errno==ENOENT )
			printf("ioctl error ENOENT %d\n",fd);
	}
	else
	{
	}
	xqry.query = UVC_GET_CUR;
	if( (r=xioctl(fd,UVCIOC_CTRL_QUERY,&xqry))<0 )
	{
		if( errno == ENOENT )
			printf("ioctl error ENOENT %d\n",fd);
	}
#else
	if ( (r=ioctl(fd,UVCIOC_CTRL_QUERY,&xqry))<0 ) {
		if ( errno==EBADF )
			printf("ioctl error EBADF\n");
		else
		if ( errno==EFAULT )
			printf("ioctl error EFAULT\n");
		else if ( errno== EINVAL )
			printf("ioctl error EINVAL\n");
		else if ( errno==ENOTTY )
			printf("ioctl error ENOTTY\n");
		else if ( errno == ENOENT )
			printf("ioctl error ENOENT\n");
		else
			printf("ioctl error! %d\n",errno);
		r=-1;	
	}
	else {
		xqry.query = UVC_GET_CUR;
		if ( (r=ioctl(fd,UVCIOC_CTRL_QUERY,&xqry))<0 ) {
			r=-1;
		}
	}
#endif
	free(xqry.data);
	return r;
}

ofxUvcCamera::ofxUvcCamera(const char* devName, int width, int height, io_method method /*= IO_METHOD_MMAP*/) :
	buffersCount(0), deviceOpened(0) 
{
	//printf("[App] %s\n", devName);
	sprintf(dev_name, "%s", devName);
	int ret;
        if (ret=open_device()) 
	{
		fprintf(stderr, "Open Device: %d\n", ret);
		throw "Open Device";
	}

	io = method;
	frame_width = width;
	frame_height = height;

	++deviceOpened;
	init_device();
	++deviceOpened;
}

ofxUvcCamera::~ofxUvcCamera()
{
	if(deviceOpened)
	{
		if(deviceOpened>=2)
			uninit_device();
		close_device();
	}
}

int ofxUvcCamera::getWidth()
{
	return frame_width;
}

int ofxUvcCamera::getHeight()
{
	return frame_height;
}

#if 0
inline void ofxUvcCamera::process_image(const void *p, int size)
{
	if (out_buf)
		fwrite(p, size, 1, stdout);

	fflush(stderr);
	fprintf(stderr, ".");
	fflush(stdout);
}
#endif

buffer* ofxUvcCamera::read_frame(void)
{
	fd_set fds;
	struct timeval tv;
	int r;

	FD_ZERO(&fds);
	FD_SET(fd, &fds);

	/* Timeout. */
	tv.tv_sec = 2;
	tv.tv_usec = 0;

	r = select(fd + 1, &fds, NULL, NULL, &tv);


	if (-1 == r) {
	//	if (EINTR == errno)
	//		continue;
	//	errno_exit("select");
		return NULL;
	}

	if (0 == r) {
		fprintf(stderr, "select timeout %d\n",fd);
	//	exit(EXIT_FAILURE);
		return NULL;
	}

	if( _read_frame() <= 0 )
	{
		fprintf(stderr," read frame =0 %d\n",fd);	
		return NULL;
	}
	/* EAGAIN - continue select loop. */

	return &buffers[(currBufferId++)%buffersCount];
}

int ofxUvcCamera::_read_frame(void)
{
        struct v4l2_buffer buf;
        unsigned int i;

        switch (io) {
        case IO_METHOD_READ:
                if (-1 == read(fd, buffers[0].start, buffers[0].length)) {
                        switch (errno) {
                        case EAGAIN:
                                return 0;

                        case EIO:
                                /* Could ignore EIO, see spec. */

                                /* fall through */

                        default:
                                //errno_exit("read");
				return -1;
                        }
                }

                //process_image(buffers[0].start, buffers[0].length);
                break;

        case IO_METHOD_MMAP:
                CLEAR(buf);

                buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                buf.memory = V4L2_MEMORY_MMAP;

                if (-1 == xioctl(fd, VIDIOC_DQBUF, &buf)) {
                        switch (errno) {
                        case EAGAIN:
                                return 0;

                        case EIO:
                                /* Could ignore EIO, see spec. */

                                /* fall through */

                        default:
                                errno_exit("VIDIOC_DQBUF");
                        }
                }

                assert(buf.index < n_buffers);

                //process_image(buffers[buf.index].start, buf.bytesused);

                if (-1 == xioctl(fd, VIDIOC_QBUF, &buf))
                        //errno_exit("VIDIOC_QBUF");
			return -2;
                break;

        case IO_METHOD_USERPTR:
                CLEAR(buf);

                buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                buf.memory = V4L2_MEMORY_USERPTR;

                if (-1 == xioctl(fd, VIDIOC_DQBUF, &buf)) {
                        switch (errno) {
                        case EAGAIN:
                                return 0;

                        case EIO:
                                /* Could ignore EIO, see spec. */

                                /* fall through */

                        default:
                                //errno_exit("VIDIOC_DQBUF");
				return -2;
                        }
                }

                for (i = 0; i < n_buffers; ++i)
                        if (buf.m.userptr == (unsigned long)buffers[i].start
                            && (unsigned long)buf.length == buffers[i].length)
                                break;

                assert(i < n_buffers);

                //process_image((void *)buf.m.userptr, buf.bytesused);

                if (-1 == xioctl(fd, VIDIOC_QBUF, &buf))
                        //errno_exit("VIDIOC_QBUF");
			return -2;
                break;
        }

        return 1;
}

#if 0
void ofxUvcCamera::mainloop(void)
{
        unsigned int count;

        count = frame_count;

        while (count-- > 0) {
                for (;;) {
                        fd_set fds;
                        struct timeval tv;
                        int r;

                        FD_ZERO(&fds);
                        FD_SET(fd, &fds);

                        /* Timeout. */
                        tv.tv_sec = 2;
                        tv.tv_usec = 0;

                        r = select(fd + 1, &fds, NULL, NULL, &tv);

                        if (-1 == r) {
                                if (EINTR == errno)
                                        continue;
                                errno_exit("select");
                        }

                        if (0 == r) {
                                fprintf(stderr, "select timeout\n");
                                exit(EXIT_FAILURE);
                        }

                        if (_read_frame())
                                break;
                        /* EAGAIN - continue select loop. */
                }
        }
}
#endif

void ofxUvcCamera::stop_capturing(void)
{
        enum v4l2_buf_type type;

        switch (io) {
        case IO_METHOD_READ:
                /* Nothing to do. */
                break;

        case IO_METHOD_MMAP:
        case IO_METHOD_USERPTR:
                type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                if (-1 == xioctl(fd, VIDIOC_STREAMOFF, &type))
                        errno_exit("VIDIOC_STREAMOFF");
                break;
        }
}

void ofxUvcCamera::start_capturing(void)
{
        unsigned int i;
        enum v4l2_buf_type type;

        switch (io) {
        case IO_METHOD_READ:
                /* Nothing to do. */
                break;

        case IO_METHOD_MMAP:
                for (i = 0; i < n_buffers; ++i) {
                        struct v4l2_buffer buf;

                        CLEAR(buf);
                        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                        buf.memory = V4L2_MEMORY_MMAP;
                        buf.index = i;

                        if (-1 == xioctl(fd, VIDIOC_QBUF, &buf))
                                errno_exit("VIDIOC_QBUF");
                }
                type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                if (-1 == xioctl(fd, VIDIOC_STREAMON, &type))
                        errno_exit("VIDIOC_STREAMON");
                break;

        case IO_METHOD_USERPTR:
                for (i = 0; i < n_buffers; ++i) {
                        struct v4l2_buffer buf;

                        CLEAR(buf);
                        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                        buf.memory = V4L2_MEMORY_USERPTR;
                        buf.index = i;
                        buf.m.userptr = (unsigned long)buffers[i].start;
                        buf.length = buffers[i].length;

                        if (-1 == xioctl(fd, VIDIOC_QBUF, &buf))
                                errno_exit("VIDIOC_QBUF");
                }
                type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                if (-1 == xioctl(fd, VIDIOC_STREAMON, &type))
                        errno_exit("VIDIOC_STREAMON");
                break;
        }
}

void ofxUvcCamera::uninit_device(void)
{
        unsigned int i;

        switch (io) {
        case IO_METHOD_READ:
                free(buffers[0].start);
                break;

        case IO_METHOD_MMAP:
                for (i = 0; i < n_buffers; ++i)
                        if (-1 == munmap(buffers[i].start, buffers[i].length))
                                errno_exit("munmap");
                break;

        case IO_METHOD_USERPTR:
                for (i = 0; i < n_buffers; ++i)
                        free(buffers[i].start);
                break;
        }

        free(buffers);
}

#define _LOCAL_BUFFER_COUNT 1
void ofxUvcCamera::init_read(unsigned int buffer_size)
{
        buffers = (buffer*)calloc(_LOCAL_BUFFER_COUNT, sizeof(*buffers));

        if (!buffers) {
                fprintf(stderr, "Out of memory\n");
                exit(EXIT_FAILURE);
        }

        buffers[0].length = buffer_size;
        buffers[0].start = malloc(buffer_size);

        if (!buffers[0].start) {
                fprintf(stderr, "Out of memory\n");
                exit(EXIT_FAILURE);
        }

	buffersCount = _LOCAL_BUFFER_COUNT;
}
#undef _LOCAL_BUFFER_COUNT

void ofxUvcCamera::init_mmap(void)
{
        struct v4l2_requestbuffers req;

        CLEAR(req);

        req.count = 4;
        req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        req.memory = V4L2_MEMORY_MMAP;

        if (-1 == xioctl(fd, VIDIOC_REQBUFS, &req)) {
                if (EINVAL == errno) {
                        fprintf(stderr, "%s does not support "
                                 "memory mapping\n", dev_name);
                        exit(EXIT_FAILURE);
                } else {
                        errno_exit("VIDIOC_REQBUFS");
                }
        }

        if (req.count < 2) {
                fprintf(stderr, "Insufficient buffer memory on %s\n",
                         dev_name);
                exit(EXIT_FAILURE);
        }

        buffers = (buffer*)calloc(req.count, sizeof(*buffers));

        if (!buffers) {
                fprintf(stderr, "Out of memory\n");
                exit(EXIT_FAILURE);
        }

        for (n_buffers = 0; n_buffers < req.count; ++n_buffers) {
                struct v4l2_buffer buf;

                CLEAR(buf);

                buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                buf.memory      = V4L2_MEMORY_MMAP;
                buf.index       = n_buffers;

                if (-1 == xioctl(fd, VIDIOC_QUERYBUF, &buf))
                        errno_exit("VIDIOC_QUERYBUF");

                buffers[n_buffers].length = buf.length;
                buffers[n_buffers].start =
                        mmap(NULL /* start anywhere */,
                              buf.length,
                              PROT_READ | PROT_WRITE /* required */,
                              MAP_SHARED /* recommended */,
                              fd, buf.m.offset);

                if (MAP_FAILED == buffers[n_buffers].start)
                        errno_exit("mmap");
        }

	buffersCount = req.count;
}

void ofxUvcCamera::init_userp(unsigned int buffer_size)
{
        struct v4l2_requestbuffers req;

        CLEAR(req);

        req.count  = 4;
        req.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        req.memory = V4L2_MEMORY_USERPTR;

        if (-1 == xioctl(fd, VIDIOC_REQBUFS, &req)) {
                if (EINVAL == errno) {
                        fprintf(stderr, "%s does not support "
                                 "user pointer i/o\n", dev_name);
                        exit(EXIT_FAILURE);
                } else {
                        errno_exit("VIDIOC_REQBUFS");
                }
        }

        buffers = (buffer*)calloc(4, sizeof(*buffers));

        if (!buffers) {
                fprintf(stderr, "Out of memory\n");
                exit(EXIT_FAILURE);
        }

        for (n_buffers = 0; n_buffers < 4; ++n_buffers) {
                buffers[n_buffers].length = buffer_size;
                buffers[n_buffers].start = malloc(buffer_size);

                if (!buffers[n_buffers].start) {
                        fprintf(stderr, "Out of memory\n");
                        exit(EXIT_FAILURE);
                }
        }

	buffersCount = req.count;
}

void ofxUvcCamera::init_device(void)
{
        struct v4l2_capability cap;
        struct v4l2_cropcap cropcap;
        struct v4l2_crop crop;
        struct v4l2_format fmt;
        unsigned int min;


        if (-1 == xioctl(fd, VIDIOC_QUERYCAP, &cap)) {
                if (EINVAL == errno) {
                        fprintf(stderr, "%s is no V4L2 device\n",
                                 dev_name);
                        exit(EXIT_FAILURE);
                } else {
                        errno_exit("VIDIOC_QUERYCAP");
                }
        }

        if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
                fprintf(stderr, "%s is no video capture device\n",
                         dev_name);
                exit(EXIT_FAILURE);
        }

        switch (io) {
        case IO_METHOD_READ:
                if (!(cap.capabilities & V4L2_CAP_READWRITE)) {
                        fprintf(stderr, "%s does not support read i/o\n",
                                 dev_name);
                        exit(EXIT_FAILURE);
                }
                break;

        case IO_METHOD_MMAP:
        case IO_METHOD_USERPTR:
                if (!(cap.capabilities & V4L2_CAP_STREAMING)) {
                        fprintf(stderr, "%s does not support streaming i/o\n",
                                 dev_name);
                        exit(EXIT_FAILURE);
                }
                break;
        }


        /* Select video input, video standard and tune here. */


        CLEAR(cropcap);

        cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

        if (0 == xioctl(fd, VIDIOC_CROPCAP, &cropcap)) {
                crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                crop.c = cropcap.defrect; /* reset to default */

                if (-1 == xioctl(fd, VIDIOC_S_CROP, &crop)) {
                        switch (errno) {
                        case EINVAL:
                                /* Cropping not supported. */
                                break;
                        default:
                                /* Errors ignored. */
                                break;
                        }
                }
        } else {
                /* Errors ignored. */
        }


        CLEAR(fmt);

        fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	//if (force_format) {
	if (frame_width!=0 && frame_height!=0) {
                fmt.fmt.pix.width       = frame_width;
                fmt.fmt.pix.height      = frame_height;
                fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
                fmt.fmt.pix.field       = V4L2_FIELD_INTERLACED;

                if (-1 == xioctl(fd, VIDIOC_S_FMT, &fmt))
                        errno_exit("VIDIOC_S_FMT");

                /* Note VIDIOC_S_FMT may change width and height. */
        } else {
                /* Preserve original settings as set by v4l2-ctl for example */
                if (-1 == xioctl(fd, VIDIOC_G_FMT, &fmt))
                        errno_exit("VIDIOC_G_FMT");

		frame_width = fmt.fmt.pix.width;
		frame_height = fmt.fmt.pix.height;
        }

        /* Buggy driver paranoia. */
        min = fmt.fmt.pix.width* 2;
        if (fmt.fmt.pix.bytesperline < min)
                fmt.fmt.pix.bytesperline = min;
        min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
        if (fmt.fmt.pix.sizeimage < min)
                fmt.fmt.pix.sizeimage = min;

        switch (io) {
        case IO_METHOD_READ:
                init_read(fmt.fmt.pix.sizeimage);
                break;

        case IO_METHOD_MMAP:
                init_mmap();
                break;

        case IO_METHOD_USERPTR:
                init_userp(fmt.fmt.pix.sizeimage);
                break;
        }
}

void ofxUvcCamera::close_device(void)
{
        if (-1 == close(fd))
                errno_exit("close");

        fd = -1;
}

int ofxUvcCamera::open_device(void)
{
        struct stat st;

        if (-1 == stat(dev_name, &st)) {
                //fprintf(stderr, "Cannot identify '%s': %d, %s\n", 
		//	dev_name, errno, strerror(errno));
		//exit(EXIT_FAILURE);
		return -1;
        }

        if (!S_ISCHR(st.st_mode)) {
		//fprintf(stderr, "%s is no device\n", dev_name);
		//exit(EXIT_FAILURE);
		return -2;
        }

        fd = open(dev_name, O_RDWR /* required */ | O_NONBLOCK, 0);

        if (-1 == fd) {
		//fprintf(stderr, "Cannot open '%s': %d, %s\n",
		//	dev_name, errno, strerror(errno));
		//exit(EXIT_FAILURE);
		return -3;
        }

	return 0;
}

